import torch
import torch.nn as nn
from torch.nn.parameter import Parameter
import math


"""
    Implements the policy network for candidates
"""


class GraphConvolution(nn.Module):
    """
    Simple GCN layer, similar to https://arxiv.org/abs/1609.02907
    """

    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input, adj):
        support = torch.mm(input, self.weight)
        output = torch.mm(adj, support)
        if self.bias is not None:
            return output + self.bias
        else:
            return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


class CandidatePolicy(nn.Module):
    def __init__(self, message_dim=128, hidden_dim=128, pref_dim=128):
        """
        :param message_dim: Dimension of message encoding
        :param hidden_dim: Dimension of hidden embedding used by memory LSTM
        :param pref_dim: Dimension of the preference vector
        """
        super(CandidatePolicy, self).__init__()

        # Initialize the variables
        self.message_dim = message_dim
        self.hidden_dim = hidden_dim
        self.pref_dim = pref_dim

        # Initialize GCN
        self.gcn1 = GraphConvolution(2, 16)
        self.act1 = nn.ELU()
        self.gcn2 = GraphConvolution(16, 32)
        self.act2 = nn.ELU()
        self.gcn3 = GraphConvolution(32, self.message_dim)
        self.pool = nn.AdaptiveMaxPool2d((1, self.message_dim))

        # Initialize the memory LSTM
        self.memory_lstm = nn.LSTM(self.message_dim, self.hidden_dim)

        # Network for sending broadcast message
        self.broadcast_linear1 = nn.Linear(self.hidden_dim + self.pref_dim, self.message_dim)
        self.broadcast_act1 = nn.ELU()
        self.broadcast_linear2 = nn.Linear(self.message_dim, self.message_dim)

    def forward(self, memory, adj_mat, current_voting, pref_vector):
        """
        :param memory: Hidden state for memory LSTM
        :param adj_mat: (num_nodes, num_nodes) Binary sparse adjacency matrix
        :param current_voting: (num_nodes, 1) Current voting decision
        :param pref_vector: Preference vector for the candidate
        :return:
            memory_state: Hidden state of memory LSTM after update
            message: (1, message_dim) Message that has to be broadcasted
        """
        # Get the input for LSTM by running GCN
        lstm_input = self.act1(self.gcn1(current_voting, adj_mat))
        lstm_input = self.act2(self.gcn2(lstm_input, adj_mat))
        lstm_input = self.pool(self.gcn3(lstm_input, adj_mat).unsqueeze(0))

        # Pass it through memory LSTM
        output, memory = self.memory_lstm(lstm_input, memory)
        output = torch.cat([output.squeeze(0), pref_vector], dim=1)

        # Compute message to be broadcasted
        h = self.broadcast_act1(self.broadcast_linear1(output))
        message = self.broadcast_linear2(h)

        return memory, message
