import torch
import torch.nn as nn
from Code.util import to_gpu
from Code.GumbelSoftmax import GumbelSoftmax


"""
    Implements the communicator that uses discrete symbols
"""


class Vocabulary(nn.Module):
    """
    Implements tha embeddings that are used by words in the vocabulary
    """

    def __init__(self, num_words, embedding_dim=128):
        """
        :param num_words: Number of words in the vocabulary
        :param embedding_dim: Dimension of the embeddings
        """
        super(Vocabulary, self).__init__()

        self.num_words = num_words
        self.embedding_dim = embedding_dim

        # +2 is needed for start token, last is start token, second last is end token
        self.embeddings = nn.Embedding(self.num_words + 2, self.embedding_dim)

    def forward(self, indices):
        """
        :param indices: List of word indices in the current sentence (note that this function handles only one sentence)
        :return: sentence
            sentence: (sentence_length, embedding_dim) sentence representation
        """
        return self.embeddings(to_gpu(torch.tensor(indices)).long())


class Encoder(nn.Module):
    """
    Implements the encoder for generating a message given SOM as input
    """

    def __init__(self, vocab, message_dim=128, max_len=25):
        """
        :param vocab: Object of the Vocabulary class
        :param message_dim: Dimension of encoded message vector
        :param max_len: Maximum length of a sentence that can be generated
        """
        super(Encoder, self).__init__()

        self.vocabulary = vocab
        self.message_dim = message_dim
        self.max_len = max_len

        self.encoder_lstm = nn.LSTM(input_size=self.vocabulary.embedding_dim, hidden_size=self.message_dim)
        self.linear = nn.Linear(self.message_dim, self.vocabulary.num_words + 1)
        self.act = GumbelSoftmax()
        self.temp_linear = nn.Linear(self.message_dim, 1)

    def forward(self, init_hidden, min_temp=0.2):
        """
        :param init_hidden: (1, message_dim) Message encoding
        :param min_temp: Minimum temperature used by Gumbel Softmax
        :return: sentence
            sentence: (message_length, vocab_size)
        """
        # Get all words
        all_words = self.vocabulary(list(range(self.vocabulary.num_words + 1)))

        # Prepare the initial hidden state
        h0 = to_gpu(torch.zeros(1, 1, self.message_dim))
        c0 = init_hidden.unsqueeze(0)
        state = (h0, c0)

        # Prepare the initial input
        inp = self.vocabulary([self.vocabulary.num_words + 1]).unsqueeze(0)

        # Run the LSTM
        count = 0
        sentence = []
        while count < self.max_len:
            # Run the LSTM
            output, state = self.encoder_lstm(inp, state)
            inv_temp = torch.log(1 + torch.exp(self.temp_linear(output.squeeze(0)))) + min_temp

            # Sample using Gumbel Softmax
            output = self.linear(output.squeeze(0))  # 1 x vocab_size
            output = self.act(output, inv_temp)
            output = self.act.create_one_hot(output)    # 1 x vocab_size

            # Add word to sentences
            sentence.append(output)

            # Check for end token
            idx = torch.max(output, dim=1)[1]
            if idx[0].item() == self.vocabulary.num_words:
                break

            # Prepare the next input
            inp = torch.mm(output, all_words).unsqueeze(0)
            count += 1

        sentence = torch.cat(sentence, dim=0)
        return sentence


class Decoder(nn.Module):
    """
    Decode the obtained sentence to get meaning of the sentence
    """

    def __init__(self, vocab, message_dim=128):
        """
        :param vocab: Object of the Vocabulary class
        :param message_dim: Dimension of output message vector
        """
        super(Decoder, self).__init__()

        self.vocabulary = vocab
        self.message_dim = message_dim

        self.decoder_lstm = nn.LSTM(self.vocabulary.embedding_dim, self.message_dim)

    def forward(self, sentences):
        """
        :param sentences: List of (message_length, vocab_size) Sentence one-hot encoding (or Gumbel)
        :return:
            messages: (num_messages, message_dim) decoded messages
        """
        # Get all words
        all_words = self.vocabulary(list(range(self.vocabulary.num_words + 1))).detach()

        # Decode the messages
        bs = len(sentences)
        messages = []
        for i in range(bs):
            state = (to_gpu(torch.zeros(1, 1, self.message_dim)), to_gpu(torch.zeros(1, 1, self.message_dim)))
            s = torch.mm(sentences[i], all_words).unsqueeze(1)  # seq_len x 1 x embedding_dim
            output, _ = self.decoder_lstm(s, state)
            messages.append(output[-1, 0, :].unsqueeze(0))

        messages = torch.cat(messages, dim=0)   # bs x message_dim

        return messages
