import  numpy as np
import os
from scipy.io import mmread
from scipy.stats import chi2
import networkx as nx


"""
    Implements various graph generators
"""


def sigmoid(z):
    return 1 / (1 + np.exp(-z))


def gen_embeds(n, d, sigma=None):
    """
    :param n: Number of nodes
    :param d: Dimensionality of embeddings
    :param sigma: Standard deviation for normal distribution
    :return: embeddings
        embeddings: (n, d) embeddings matrix
    """
    if sigma is None:
        sigma = 1 / np.sqrt(d)
    return np.random.normal(scale=sigma, size=(n, d))


def dot_product_similarity(X, Y):
    """
    Computes the dot product similarity between the nodes
    :param X: (n, d) embeddings
    :param Y: (n, d) embeddings
    :return: similarity
        similarity: (n, n) similarity matrix where each entry is in the range [0, 1]
    """
    return sigmoid(np.matmul(X, Y.T))


def sample_graph(X, Y, use_activities=True, sim_fn=dot_product_similarity):
    """
    :param X: (n, d) embedding matrix
    :param Y: (n, d) embedding matrix
    :param use_activities: Whether to use activity parameter while generating the graph
    :param sim_fn: Similarity function used for computing similarity
    :return:
        sim_matrix: (n, n) similarity matrix
        adj_matrix: (n, n) sampled adjacency matrix
    """
    n = X.shape[0]

    if use_activities:
        theta1 = 0.5 + 0.5 * np.random.random(size=(n // 10, 1))
        theta2 = 0.5 * np.random.random(size=(n // 10, 1))
        theta3 = 0.5 * np.random.random(size=(n - n // 10, 2))
        thetas = np.concatenate([np.concatenate([theta2, theta1], axis=1), theta3], axis=0)
        sim_matrix = np.matmul(thetas[:, 0].reshape((-1, 1)), thetas[:, 1].reshape((1, -1))) * sim_fn(X, Y)
    else:
        sim_matrix = sim_fn(X, Y)

    adj_matrix = (np.random.random(size=(n, n)) <= sim_matrix).astype(float)

    # Add a self loop
    adj_matrix = adj_matrix - np.diag(np.diag(adj_matrix))

    return sim_matrix, adj_matrix.astype(int)


def sample_euclidean_graph(X, sparsity=0.05):
    """
    :param X: (n, d) Node Embeddings, must have variance 1/d and independent components for sparsity to work
    :param sparsity: Probability of observing an edge
    :return:
        adj_matrix: (n, n) sampled adjacency matrix
    """
    n, d = X.shape

    # Compute the threshold
    threshold = 2 * chi2.isf(1 - sparsity, df=d) / d

    # Compute the distances
    XY = np.matmul(X, X.T)
    XX = np.tile(np.diag(np.matmul(X, X.T)).reshape((1, -1)), (n, 1))
    distances = XX + XX.T - 2 * XY

    return (distances <= threshold).astype(float) - np.eye(n)


def read_wikivote(pref_dim, data_dir='../Data/soc-wiki-Vote/'):
    """
    :param data_dir: Path to soc-wiki-Vote folder
    :param pref_dim: Dimension of preference vector
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    with open(os.path.join(data_dir, 'soc-wiki-Vote.mtx')) as f:
        # Read the lines
        lines = [line for line in f]

    # Initialize the adjacency matrix
    num_nodes = int(lines[1].split(' ')[0])
    adj_mat = np.zeros((num_nodes, num_nodes))

    # Read the edges
    lines = lines[2:]
    for line in lines:
        n1, n2 = line.split(' ')
        n1 = int(n1) - 1
        n2 = int(n2) - 1
        adj_mat[n1, n2] = 1
        adj_mat[n2, n1] = 1

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs / np.tile(np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds


def read_netsci_collab(pref_dim, data_dir='../Data/ca-netscience/'):
    """
    :param data_dir: Path to ca-netscience folder
    :param pref_dim: Dimension of preference vector
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    # Read the adjacency matrix
    adj_mat = mmread(os.path.join(data_dir, 'ca-netscience.mtx')).todense()

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs / np.tile(np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds


def read_karate(pref_dim, data_dir='../Data/soc-karate/'):
    """
    :param data_dir: Path to soc-karate folder
    :param pref_dim: Dimension of preference vector
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    # Read the adjacency matrix
    adj_mat = mmread(os.path.join(data_dir, 'soc-karate.mtx')).todense()

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs / np.tile(np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds


def sample_sbm(num_nodes, pref_dim, sparsity=0.1, pi=None, B=None):
    """
    :param num_nodes: Number of nodes in the network
    :param pref_dim: Dimension of preference vector
    :param sparsity: Sparsity parameter
    :param pi: (1, K) Community membership probability prior
    :param B: (K, K) Block matrix
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    if pi is None:
        pi = np.ones(shape=(1, 2)) / 2.0

    if B is None:
        B = np.random.random(size=(2, 2)) + np.eye(2)
        B = B / B.max()

    K = B.shape[0]
    Z = np.random.choice(K, num_nodes, p=pi.squeeze())
    Z = np.eye(K)[Z, :]

    # Get the adjacency matrix
    adj_prob = np.matmul(np.matmul(Z, B), Z.T)
    adj_mat = (np.random.random(size=adj_prob.shape) <= sparsity * adj_prob).astype(float)
    adj_mat = np.triu(adj_mat, 1)
    adj_mat = adj_mat + adj_mat.T

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs / np.tile(1e-6 + np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds


def planted_partition_generate(num_nodes, num_comm, pref_dim=None, lambda_in=30, lambda_out=1):
    """
    :param num_nodes: Number of nodes
    :param num_comm: Number of communities
    :param pref_dim: Preference vector dimension
    :param lambda_in: Expected in-degree
    :param lambda_out: Expected out-degree
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    if pref_dim is None:
        pref_dim = num_comm

    # Sample the adjacency matrix
    p_in = lambda_in / num_nodes
    p_out = lambda_out / num_nodes
    adj_mat = nx.planted_partition_graph(num_comm, num_nodes // num_comm, p_in, p_out)
    adj_mat = nx.to_numpy_matrix(adj_mat)

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs / np.tile(1e-6 + np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds


def read_polbooks(pref_dim, data_dir='../Data/polbooks/'):
    """
    :param data_dir: Path to polbooks folder
    :param pref_dim: Dimension of preference vector
    :return: adj_mat, embeds
        adj_mat: (num_nodes, num_nodes) adjacency matrix
        embeds: (num_nodes, pref_dim) node embeddings
    """
    # Read the adjacency matrix
    adj_mat = mmread(os.path.join(data_dir, 'polbooks.mtx')).todense()

    # Get the node embeddings
    eig, vecs = np.linalg.eig(adj_mat)
    vecs = np.real(vecs[:, :pref_dim])
    embeds = vecs # / np.tile(np.linalg.norm(vecs, axis=1).reshape((-1, 1)), (1, pref_dim))

    return adj_mat, embeds
