import torch
from Code.util import to_gpu


"""
    Implements a population node from the network
"""


class Node(object):

    def __init__(self, policy_net, encoder, decoder, preference_vector, memory_dim=128):
        """
        :param policy_net: Instance of NodePolicy network
        :param encoder: Encoder network
        :param decoder: Decoder network
        :param preference_vector: (1, pref_dim) Initial preference vector for the node
        :param memory_dim: Dimension of memory vector used by the policy network
        """
        self.memory_dim = memory_dim

        # Initialize the networks
        self.policy_net = policy_net
        self.encoder = encoder
        self.decoder = decoder

        # Initialize the neighbors and messages
        self.neighbors = []
        self.messages = []
        self.new_messages = []

        # Initialize the preference vector
        self.preference_vector = to_gpu(torch.from_numpy(preference_vector).float())

        # Initialize the memory
        self.memory = (to_gpu(torch.zeros(1, 1, self.memory_dim + self.preference_vector.size(1))),
                       torch.cat([to_gpu(torch.zeros(1, self.memory_dim)), self.preference_vector], dim=1).unsqueeze(0))

    def reset_node(self):
        self.messages = []
        self.new_messages = []
        self.memory = (to_gpu(torch.zeros(1, 1, self.memory_dim + self.preference_vector.size(1))),
                       torch.cat([to_gpu(torch.zeros(1, self.memory_dim)), self.preference_vector], dim=1).unsqueeze(0))

    def reset_messages(self):
        self.messages = self.new_messages[:]
        self.new_messages = []

    def add_message(self, message):
        """
        :param message: (message_length, vocab_size) Represents a received message
        """
        self.new_messages.append(message)

    def add_neighbor(self, node):
        """
        :param node: Instance of node object that is a neighbor
        """
        self.neighbors.append(node)

    def broadcast(self, message):
        """
        :param message: List of words representing the message
        """
        for neighbor in self.neighbors:
            neighbor.add_message(message)

    def process_messages(self, return_message=False):
        """
        :param return_message: Whether to return the discrete message broadcasted at current step for logging.
        :return:
            message: Discrete message
        """
        if len(self.messages) > 0:
            # Run the policy network
            messages = self.decoder(self.messages)
            new_pref, pref_update, message, memory = self.policy_net(messages, self.memory, self.preference_vector)

            # Update the memory
            self.memory = memory

            # Perform updates to preference vector
            self.preference_vector = (1 - pref_update) * self.preference_vector + pref_update * new_pref

            # Broadcast the message
            message = self.encoder(message)
            self.broadcast(message)

            if return_message:
                return torch.argmax(message, dim=1).tolist()

    def get_dist(self, candidate1, candidate2):
        """
        :param candidate1: (1, pref_dim) Preference vector for first candidate
        :param candidate2: (1, pref_dim) Preference vector for second candidate
        :return: distances
            distances: (1, 2) Distance of node from candidate 1 and 2 based on preference vector
        """
        distances = torch.cat([((self.preference_vector - candidate1) ** 2).mean().unsqueeze(0).unsqueeze(0),
                               ((self.preference_vector - candidate2) ** 2).mean().unsqueeze(0).unsqueeze(0)],
                              dim=1)
        return distances
