import torch
import torch.nn as nn
from Code.util import to_gpu


"""
    Implements the Policy Network that runs on each node. Note that the same network is shared by all nodes.
"""


class NodePolicy(nn.Module):
    def __init__(self, message_dim=128, hidden_dim=128, pref_dim=128, learn_follow_temp=False, inv_temp_follow=2.0,
                 element_wise_pref_update=False):
        """
        :param message_dim: Dimension of message to be generated. Input dimension of message should be same
        :param hidden_dim: Hidden dimension used by memory LSTM
        :param pref_dim: Dimension of the preference vector
        :param learn_follow_temp: Whether to learn the inverse temperature for following
        :param inv_temp_follow: Initial inverse temperature or the parameter governing it
        :param element_wise_pref_update: Whether to perform elementwise sigmoid filtering for updating preferences
        """
        super(NodePolicy, self).__init__()

        # Network for combining messages
        self.message_dim = message_dim
        self.pref_dim = pref_dim
        self.message_linear1 = nn.Linear(self.message_dim + self.pref_dim, self.message_dim)
        self.message_act1 = nn.ELU()
        self.message_linear2 = nn.Linear(self.message_dim, self.message_dim)
        self.message_act2 = nn.ELU()
        self.message_pooling = nn.AdaptiveMaxPool2d(output_size=(1, self.message_dim))

        # Network for extracting meaning
        self.hidden_dim = hidden_dim
        self.memory_lstm = nn.LSTM(self.message_dim, self.hidden_dim + self.pref_dim)

        # Network for updating preference vector
        self.pref_linear1 = nn.Linear(self.hidden_dim + self.pref_dim, self.pref_dim)
        self.pref_act1 = nn.ELU()
        self.pref_linear2 = nn.Linear(self.pref_dim, self.pref_dim)
        if element_wise_pref_update:
            self.pref_update = nn.Linear(self.hidden_dim + self.pref_dim, self.pref_dim)
        else:
            self.pref_update = nn.Linear(self.hidden_dim + self.pref_dim, 1)
        self.update_act = nn.Sigmoid()

        # Network for sending broadcast message
        self.broadcast_linear1 = nn.Linear(self.hidden_dim + self.pref_dim, self.message_dim)
        self.broadcast_act1 = nn.ELU()
        self.broadcast_linear2 = nn.Linear(self.message_dim, self.message_dim)

        # Set up the follow temperature parameter
        if learn_follow_temp:
            self.inv_temp_follow = nn.Parameter(to_gpu(torch.tensor([inv_temp_follow])))
        else:
            self.inv_temp_follow = to_gpu(torch.tensor([inv_temp_follow]))

    def forward(self, messages, memory, pref_vector):
        """
        :param messages: (num_messages, message_dim) Messages obtained from the neighbors
        :param memory: Memory of the node to be used as hidden state for memory LSTM
        :param pref_vector: (1, pref_dim) Current preference vector of the node
        :return:
            pref_new: (1, pref_dim) New preference vector
            pref_update: Sigmoid giving update weight
            message: (1, message_dim) Message to be sent by this node
            memory: Updated value of memory state
        """
        # Combine the messages to obtain an embedding
        num_messages = messages.size(0)
        messages = torch.cat([messages, pref_vector.expand(num_messages, self.pref_dim)], dim=1)
        messages = self.message_act1(self.message_linear1(messages))
        messages = self.message_act2(self.message_linear2(messages)).unsqueeze(0)
        message = self.message_pooling(messages)

        # Extract meaning from the message
        message, memory = self.memory_lstm(message, memory)
        message = message.squeeze(0)

        # Decide whether to update preference vector and generate pref_delta, the change in preference vector
        h1 = self.pref_act1(self.pref_linear1(message))
        pref_new = self.pref_linear2(h1)
        pref_update = 0.1 * self.update_act(self.pref_update(message))

        # Decide whether to send a broadcast message and generate the message
        h2 = self.broadcast_act1(self.broadcast_linear1(message))
        message = self.broadcast_linear2(h2)

        return pref_new, pref_update, message, memory
