from Code.CandidatePolicy import CandidatePolicy
from Code.NodePolicy import NodePolicy
from Code.Candidate import Candidate
from Code.Node import Node
import Code.GraphGenerator as gg
from Code.Communicator import Encoder, Decoder, Vocabulary
from Code.Voting import Voting
import Code.util as util
import Code.Config as config
import numpy as np
import torch
import gc
import os
from sklearn.manifold import TSNE
import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt



"""
    Analysis the language learned by the network
"""

# Load the configuration
config.EXP_NAME = 'Exp_2019-01-04 10:53:23.784927'
config.read_config()
config.print_config()
config.MODELS_DIR = os.path.join(config.ROOT_DIR, 'saved_models')
config.NUM_EPISODES = 100


# Set up the voting temperature parameters
config.INV_TEMP_VOTE = util.to_gpu(torch.tensor([config.INV_TEMP_VOTE]))


# Initialize the communicator
vocab = util.to_gpu(Vocabulary(config.NUM_WORDS, config.EMBEDDING_DIM))
encoder = util.to_gpu(Encoder(vocab, config.MESSAGE_DIM, config.MAX_LENGTH))
decoder = util.to_gpu(Decoder(vocab, config.MESSAGE_DIM))


# Initialize the node policy network
node_policy = util.to_gpu(NodePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_NODE, config.PREF_DIM,
                                     config.LEARN_FOLLOW_TEMP, config.INV_TEMP_FOLLOW,
                                     config.ELEMENT_SIGMOID_PREF_UPDATE))


# Initialize the candidate policy networks
candidate_policy1 = util.to_gpu(CandidatePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_CANDIDATE, config.PREF_DIM))
candidate_policy2 = util.to_gpu(CandidatePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_CANDIDATE, config.PREF_DIM))


# Initialize the voting network
voting = Voting()


# Restore or initialize weights
encoder.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_encoder')))
decoder.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_decoder')))
vocab.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_vocab')))
node_policy.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_node')))
candidate_policy1.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_c1')))
candidate_policy2.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_c2')))


# Plot the vocabulary
tsne_plot = TSNE().fit_transform(vocab.embeddings.weight.detach().numpy())
plt.plot(tsne_plot[:, 0], tsne_plot[:, 1], 'rx')
plt.savefig(os.path.join(config.ROOT_DIR, 'vocab.pdf'))


# Initialize the discourses file
dis_file = open(os.path.join(config.ROOT_DIR, 'discourses.txt'), 'w')


# Start testing
for epoch in range(config.NUM_EPISODES):

    dis_file.write('*****************************\n')

    # Get the network
    init_pref_vectors = gg.gen_embeds(config.NUM_NODES, config.PREF_DIM)
    adj_mat = gg.sample_euclidean_graph(init_pref_vectors)
    # adj_mat, init_pref_vectors = gg.planted_partition_generate(config.NUM_NODES, 4, config.PREF_DIM)
    adj_mat = util.to_gpu(torch.from_numpy(adj_mat).float())
    c1_pref_vector = np.random.normal(size=(1, config.PREF_DIM))
    # c2_pref_vector = np.random.normal(size=(1, config.PREF_DIM))
    c2_pref_vector = -c1_pref_vector

    # Initialize the nodes in the network
    nodes = [Node(node_policy, encoder, decoder, init_pref_vectors[i, :].reshape((1, -1)), config.HIDDEN_DIM_NODE)
             for i in range(config.NUM_NODES)]

    # Initialize the candidates in the network
    candidate1 = Candidate(candidate_policy1, encoder, decoder, c1_pref_vector, config.HIDDEN_DIM_CANDIDATE)
    candidate2 = Candidate(candidate_policy2, encoder, decoder, c2_pref_vector, config.HIDDEN_DIM_CANDIDATE)
    candidate_vectors = torch.cat([candidate1.preference_vector.detach(),
                                   candidate2.preference_vector.detach()], dim=0)

    # Join the nodes
    for i in range(config.NUM_NODES):
        for j in range(i):
            if adj_mat[i, j] == 1:
                nodes[j].add_neighbor(nodes[i])
                nodes[i].add_neighbor(nodes[j])

    # Get the episodes
    layout = None
    for t in range(config.EPISODE_LENGTH):

        dis_file.write('################################\n')

        # Get the followers for candidates
        candidate1.reset_followers()
        candidate2.reset_followers()
        idx = 0
        following = util.to_gpu(torch.zeros(config.NUM_NODES, 2))
        for node in nodes:
            following[idx, :] = voting(node.get_dist(candidate1.preference_vector.detach(),
                                                     candidate2.preference_vector.detach()),
                                       node_policy.inv_temp_follow)
            if following[idx, 0].detach() > following[idx, 1].detach():
                candidate1.add_follower(node)
            else:
                candidate2.add_follower(node)
            idx += 1

        # Run policy for candidates
        message_c1 = candidate1.take_action(adj_mat, following, return_message=True)
        message_c2 = candidate2.take_action(adj_mat, following, return_message=True)
        dis_file.write('C1\t' + '\t'.join([str(x) for x in message_c1]) + '\n')
        dis_file.write('C2\t' + '\t'.join([str(x) for x in message_c2]) + '\n')

        # Reset the messages in nodes
        for node in nodes:
            node.reset_messages()

        # Process the messages at nodes
        idx = 0
        for node in nodes:
            message = node.process_messages(return_message=True)
            dis_file.write(str(idx) + '\t' + str(int(following[idx, 0].detach() > following[idx, 1].detach())) +
                           '\t' + '\t'.join([str(x) for x in message]) + '\n')
            idx += 1

    gc.collect()

dis_file.close()
