from Code.CandidatePolicy import CandidatePolicy
from Code.NodePolicy import NodePolicy
from Code.Candidate import Candidate
from Code.Node import Node
import Code.GraphGenerator as gg
from Code.Communicator import Encoder, Decoder, Vocabulary
from Code.Voting import Voting
import Code.util as util
import Code.Config as config
import numpy as np
import torch
from torch.optim import Adam
import gc
import sys
import os



"""
    Trains the networks
"""


# Display the configuration
config.print_config()

if not config.RESTORE_FROM_CONFIG:
    sys.stdout = open(os.path.join(config.ROOT_DIR, 'config.txt'), 'w')
    config.print_config()

if config.LOG_FILE is not None:
    sys.stdout = open(config.LOG_FILE, 'w+')
else:
    sys.stdout = sys.__stdout__


# Set up the voting temperature parameters
config.INV_TEMP_VOTE = util.to_gpu(torch.tensor([config.INV_TEMP_VOTE]))


# Decide whether to use TSNE to plotting
if config.PREF_DIM > 2:
    USE_TSNE = True
else:
    USE_TSNE = False


# Initialize the communicator
vocab = util.to_gpu(Vocabulary(config.NUM_WORDS, config.EMBEDDING_DIM))
encoder = util.to_gpu(Encoder(vocab, config.MESSAGE_DIM, config.MAX_LENGTH))
decoder = util.to_gpu(Decoder(vocab, config.MESSAGE_DIM))


# Initialize the node policy network
node_policy = util.to_gpu(NodePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_NODE, config.PREF_DIM,
                                     config.LEARN_FOLLOW_TEMP, config.INV_TEMP_FOLLOW,
                                     config.ELEMENT_SIGMOID_PREF_UPDATE))


# Initialize the candidate policy networks
candidate_policy1 = util.to_gpu(CandidatePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_CANDIDATE, config.PREF_DIM))
candidate_policy2 = util.to_gpu(CandidatePolicy(config.MESSAGE_DIM, config.HIDDEN_DIM_CANDIDATE, config.PREF_DIM))


# Initialize the voting network
voting = Voting()


# Restore or initialize weights
if config.RESTORE_FROM_CONFIG:
    encoder.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_encoder')))
    decoder.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_decoder')))
    vocab.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_vocab')))
    node_policy.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_node')))
    candidate_policy1.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_c1')))
    candidate_policy2.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(config.MODEL_NUM) + '_c2')))
else:
    for param in encoder.parameters():
        if len(param.size()) >= 2:
            torch.nn.init.xavier_uniform_(param)
    for param in decoder.parameters():
        if len(param.size()) >= 2:
            torch.nn.init.xavier_uniform_(param)
    for param in node_policy.parameters():
        if len(param.size()) >= 2:
            torch.nn.init.xavier_uniform_(param)
    for param in candidate_policy1.parameters():
        if len(param.size()) >= 2:
            torch.nn.init.xavier_uniform_(param)
    for param in candidate_policy2.parameters():
        if len(param.size()) >= 2:
            torch.nn.init.xavier_uniform_(param)
    torch.nn.init.uniform_(node_policy.pref_update.bias, a=-3, b=-2)


# Set up the optimizers
if config.NODE_EDIT_COMMUNICATOR:
    optim_node = Adam(list(encoder.parameters()) + list(decoder.parameters()) + list(vocab.parameters()) +
                      list(node_policy.parameters()), lr=config.LEARNING_RATE)
else:
    optim_node = Adam(list(node_policy.parameters()), lr=config.LEARNING_RATE)

if config.CANDIDATE_EDIT_COMMUNICATOR:
    optim_c1 = Adam(list(encoder.parameters()) + list(decoder.parameters()) + list(vocab.parameters()) +
                    list(candidate_policy1.parameters()), lr=config.LEARNING_RATE)
    optim_c2 = Adam(list(encoder.parameters()) + list(decoder.parameters()) + list(vocab.parameters()) +
                    list(candidate_policy2.parameters()), lr=config.LEARNING_RATE)
else:
    optim_c1 = Adam(list(candidate_policy1.parameters()), lr=config.LEARNING_RATE)
    optim_c2 = Adam(list(candidate_policy2.parameters()), lr=config.LEARNING_RATE)


# Get the network
# adj_mat, init_pref_vectors = gg.read_wikivote(config.PREF_DIM, '../Data/soc-wiki-Vote/')
# adj_mat, init_pref_vectors = gg.read_netsci_collab(config.PREF_DIM, '../Data/ca-netscience/')
adj_mat, init_pref_vectors = gg.read_polbooks(config.PREF_DIM, '../Data/polbooks/')
assert config.NUM_NODES == adj_mat.shape[0], "Invalid configuration specification for number of nodes"
adj_mat = util.to_gpu(torch.from_numpy(adj_mat).float())
nodes = [Node(node_policy, encoder, decoder, init_pref_vectors[i, :].reshape((1, -1)), config.HIDDEN_DIM_NODE)
             for i in range(config.NUM_NODES)]
for i in range(config.NUM_NODES):
    for j in range(i):
        if adj_mat[i, j] == 1:
            nodes[j].add_neighbor(nodes[i])
            nodes[i].add_neighbor(nodes[j])


# Start training
c1_win = 0
c2_win = 0
c1_win_temp = 0
c2_win_temp = 0
draw = 0
for epoch in range(config.NUM_EPISODES):

    # Get preference vector for candidates
    c1_pref_vector = np.random.normal(size=(1, config.PREF_DIM))
    # c2_pref_vector = np.random.normal(size=(1, config.PREF_DIM))
    c2_pref_vector = -c1_pref_vector

    # Initialize the nodes in the network
    for idx in range(config.NUM_NODES):
        nodes[idx].preference_vector = util.to_gpu(torch.from_numpy(init_pref_vectors[idx, :].reshape((1, -1))).float())
        nodes[idx].reset_node()

    # Initialize the candidates in the network
    candidate1 = Candidate(candidate_policy1, encoder, decoder, c1_pref_vector, config.HIDDEN_DIM_CANDIDATE)
    candidate2 = Candidate(candidate_policy2, encoder, decoder, c2_pref_vector, config.HIDDEN_DIM_CANDIDATE)
    candidate_vectors = torch.cat([candidate1.preference_vector.detach(),
                                   candidate2.preference_vector.detach()], dim=0)

    # Reset the optimizers
    optim_node.zero_grad()
    optim_c1.zero_grad()
    optim_c2.zero_grad()

    # Get the episodes
    layout = None
    for t in range(config.EPISODE_LENGTH):

        # Get the followers for candidates
        candidate1.reset_followers()
        candidate2.reset_followers()
        idx = 0
        following = util.to_gpu(torch.zeros(config.NUM_NODES, 2))
        for node in nodes:
            following[idx, :] = voting(node.get_dist(candidate1.preference_vector.detach(),
                                                     candidate2.preference_vector.detach()),
                                       node_policy.inv_temp_follow)
            if following[idx, 0].detach() > following[idx, 1].detach():
                candidate1.add_follower(node)
            else:
                candidate2.add_follower(node)
            idx += 1

        # Run policy for candidates
        candidate1.take_action(adj_mat, following)
        candidate2.take_action(adj_mat, following)

        # Reset the messages in nodes
        for node in nodes:
            node.reset_messages()

        # Process the messages at nodes
        for node in nodes:
            node.process_messages()

        # Show the result of voting
        if (epoch + 1) % config.DISPLAY_STEP == 0:
            follow_max = torch.max(following, dim=1)[1].unsqueeze(1).cpu().numpy()
            if t == 0:
                layout = util.visualize_graph(adj_mat.cpu().numpy(), follow_max, directed=False, epoch=epoch, t=t,
                                              save_fig=True, destination_dir=config.PLOTS_DIR)
            else:
                util.visualize_graph(adj_mat.cpu().numpy(), follow_max, layout, directed=False,
                                     epoch=epoch, t=t, save_fig=True, destination_dir=config.PLOTS_DIR)


            pref_vectors = util.to_gpu(torch.zeros(config.NUM_NODES, config.PREF_DIM))
            idx = 0
            for node in nodes:
                pref_vectors[idx, :] = node.preference_vector
                idx += 1
            util.plot_preferences(pref_vectors.detach().cpu().numpy(),
                                  candidate1.preference_vector.detach().cpu().numpy(),
                                  candidate2.preference_vector.detach().cpu().numpy(),
                                  follow_max, epoch, t, config.IMAGES_DIR, use_tsne=USE_TSNE)

    # Get the preference vectors
    pref_vectors = util.to_gpu(torch.zeros(config.NUM_NODES, config.PREF_DIM))
    idx = 0
    for node in nodes:
        pref_vectors[idx, :] = node.preference_vector
        idx += 1

    # Get the current distances and perform voting
    current_distances = torch.cat([((pref_vectors -
                     candidate_vectors[0, :].expand(config.NUM_NODES, config.PREF_DIM)) ** 2).mean(dim=1).unsqueeze(1),
                     ((pref_vectors -
                     candidate_vectors[1, :].expand(config.NUM_NODES, config.PREF_DIM)) ** 2).mean(dim=1).unsqueeze(1)],
                                  dim=1)
    current_voting = voting(current_distances, config.INV_TEMP_VOTE)
    vote_sum = current_voting.sum(dim=0)
    if vote_sum[0].item() > vote_sum[1].item():
        c1_win += 1
        c1_win_temp += 1
        if config.LOSS_TYPE == 'KingLoss':
            winner = candidate_vectors[0, :].unsqueeze(0).expand_as(pref_vectors)
    elif vote_sum[0].item() < vote_sum[1].item():
        c2_win += 1
        c2_win_temp += 1
        if config.LOSS_TYPE == 'KingLoss':
            winner = candidate_vectors[1, :].unsqueeze(0).expand_as(pref_vectors)
    else:
        draw += 1
        if config.LOSS_TYPE == 'KingLoss':
            winner = torch.mm(current_voting.detach(), candidate_vectors)

    if config.LOSS_TYPE == 'SheepLoss':
        winner = torch.mm(current_voting.detach(), candidate_vectors)

    # Get final result based rewards for nodes
    mean_dist_winner = ((pref_vectors - winner) ** 2).mean()
    node_reward = -mean_dist_winner

    # Get the rewards for candidates
    c1_reward = vote_sum[0]
    c2_reward = vote_sum[1]

    # Compute the gradients and take the step
    temp = np.random.random()
    if temp <= 1.0/3:
        (-node_reward).backward()
        optim_node.step()
    elif temp <= 2.0/3:
        if config.PREFERRED is None or config.PREFERRED == 'C1':
            (-c1_reward).backward()
        elif config.PREFERRED == 'C2':
            c1_reward.backward()
        optim_c1.step()
    else:
        if config.PREFERRED is None or config.PREFERRED == 'C2':
            (-c2_reward).backward()
        elif config.PREFERRED == 'C1':
            c2_reward.backward()
        optim_c2.step()

    # Show the cost
    if (epoch + 1) % config.DISPLAY_STEP == 0:
        print('Epoch:', epoch + 1,
              'Node: {0:3.5f}'.format(node_reward.item()),
              'C1: {0:.2f}'.format(c1_reward.item()),
              'C2: {0:.2f}'.format(c2_reward.item()),
              'C1 Win: {0:.4f}'.format(c1_win / (epoch + 1)),
              'C2 Win: {0:.4f}'.format(c2_win / (epoch + 1)),
              'Draw: {0:.4f}'.format(draw / (epoch + 1)),
              '[gC1]: {0:.8f}'.format(util.compute_grad(candidate_policy1)),
              '[gC2]: {0:.8f}'.format(util.compute_grad(candidate_policy2)),
              '[gN]: {0:.8f}'.format(util.compute_grad(node_policy)))
        sys.stdout.flush()

    if (epoch + 1) % config.SAVE_STEP == 0:
        torch.save(encoder.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_encoder'))
        torch.save(decoder.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_decoder'))
        torch.save(vocab.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_vocab'))
        torch.save(candidate_policy1.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_c1'))
        torch.save(candidate_policy2.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_c2'))
        torch.save(node_policy.state_dict(), os.path.join(config.MODELS_DIR, str(epoch + 1) + '_node'))

        if config.ALTERNATIVE_UPDATES:
            print('C1 Win:', c1_win_temp, 'C2 Win:', c2_win_temp)
            if config.PREFERRED is None:
                if c1_win_temp > c2_win_temp:
                    candidate_policy2.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(epoch + 1)+'_c1')))
                elif c1_win_temp < c2_win_temp:
                    candidate_policy1.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(epoch + 1)+'_c2')))

            elif config.PREFERRED == 'C1':
                candidate_policy2.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(epoch + 1)+'_c1')))

            elif config.PREFERRED == 'C2':
                candidate_policy1.load_state_dict(torch.load(os.path.join(config.MODELS_DIR, str(epoch + 1)+'_c2')))

            c1_win_temp = 0
            c2_win_temp = 0

    gc.collect()
