import Code.Config as config
import igraph as ig
import numpy as np
from sklearn.manifold import TSNE
import os
import matplotlib.pyplot as plt


def to_gpu(x):
    if config.USE_GPU:
        return x.cuda()
    return x


def visualize_graph(adj_mat, preferences=None, layout=None, directed=True, epoch=0, t=0, save_fig=False,
                    destination_dir='./plots'):
    """
    :param adj_mat: (num_nodes, num_nodes) Binary adjacency matrices of graph
    :param preferences: (num_nodes, 1) Node preferences
    :param layout: Layout to be used
    :param directed: Whether the network is directed or not
    :param epoch: Epoch number for naming saved file
    :param t: Episode number for naming saved file
    :param save_fig: Whether to save the figure or not
    :param destination_dir: Destination directory where plots are to be saved
    """
    num_nodes = adj_mat.shape[0]
    if preferences is None:
        preferences = np.zeros((num_nodes, 1))

    # Define colors for preferences
    colors = ['red', 'green', 'blue', 'cyan', 'magenta', 'yellow', 'black']

    # Show the graph
    if directed:
        edges = np.concatenate([temp.reshape(-1, 1) for temp in np.nonzero(adj_mat)], axis=1).tolist()
    else:
        edges = np.concatenate([temp.reshape(-1, 1) for temp in np.nonzero(np.triu(adj_mat, 1))], axis=1).tolist()
    graph = ig.Graph(directed=directed)
    graph.add_vertices(num_nodes)
    graph.add_edges(edges)
    graph.vs['cluster'] = preferences[:, 0].tolist()
    graph.vs["color"] = [colors[cluster] for cluster in graph.vs["cluster"]]

    if layout is None:
        layout = graph.layout_kamada_kawai()

    if save_fig:
        ig.plot(graph, os.path.join(destination_dir, str(epoch + 1) + '_' + str(t + 1) + '.png'), layout=layout)
    else:
        ig.plot(graph, layout=layout)

    return layout


def plot_preferences(preference_vectors, candidate1, candidate2, preferences=None, epoch=0, t=0,
                     destination_dir='./images', use_tsne=False):
    """
    :param preference_vectors: (num_nodes, pref_dim) Preference vectors
    :param candidate1: (1, pref_dim) candidate 1 preference vector
    :param candidate2: (1, pref_dim) candidate 2 preference vector
    :param preferences: (num_nodes, 1) binary node preferences
    :param epoch: Episode number
    :param t: Time step within an episode
    :param destination_dir: Path to the directory where plots will be saved
    :param use_tsne: Whether to use t-SNE or not
    """
    num_nodes = preference_vectors.shape[0]

    if use_tsne:
        preference_vectors = TSNE().fit_transform(np.concatenate([preference_vectors, candidate1, candidate2], axis=0))
        candidate1 = preference_vectors[-2, :].reshape((1, 2))
        candidate2 = preference_vectors[-1, :].reshape((1, 2))
        preference_vectors =  preference_vectors[:num_nodes, :]

    if preferences is None:
        preferences = np.zeros((num_nodes, 1))

    preferences = preferences.squeeze(1)
    plt.plot(preference_vectors[preferences == 0, 0], preference_vectors[preferences == 0, 1], 'gx')
    plt.plot(preference_vectors[preferences == 1, 0], preference_vectors[preferences == 1, 1], 'bx')
    plt.plot(candidate1[0, 0], candidate1[0, 1], 'gs')
    plt.plot(candidate2[0, 0], candidate2[0, 1], 'bs')

    if not use_tsne:
        plt.xlim((-2, 2))
        plt.ylim((-2, 2))

    plt.savefig(os.path.join(destination_dir, str(epoch + 1) + '_' + str(t + 1) + '.png'))
    plt.close()


def compute_grad(model):
    norm = 0.0
    for p in model.parameters():
        if p.grad is not None:
            norm += (p.grad.data ** 2).sum()
    return norm
